@@ -1,45 +1,56 @@
 //! livi is a library for hosting LV2 plugins in Rust.
 //! ```
-//! use livi;
-//!
-//! let mut world = livi::World::new();
-//! const MIN_BLOCK_SIZE: usize = 1;
-//! const MAX_BLOCK_SIZE: usize = 256;
+//! let world = livi::World::new();
 //! const SAMPLE_RATE: f64 = 44100.0;
-//! world
-//!     .initialize_block_length(MIN_BLOCK_SIZE, MAX_BLOCK_SIZE)
-//!     .unwrap();
+//! let features = world.build_features(livi::FeaturesBuilder::default());
+//! let worker_manager = features.worker_manager().clone();
 //! let plugin = world
+//!     // This is the URI for mda EPiano. You can use the `lv2ls` command line
+//!     // utility to see all available LV2 plugins.
 //!     .plugin_by_uri("http://drobilla.net/plugins/mda/EPiano")
 //!     .expect("Plugin not found.");
 //! let mut instance = unsafe {
 //!     plugin
-//!         .instantiate(SAMPLE_RATE)
+//!         .instantiate(features.clone(), SAMPLE_RATE)
 //!         .expect("Could not instantiate plugin.")
 //! };
+//!
+//! // Where midi events will be read from.
 //! let input = {
-//!     let mut s = livi::event::LV2AtomSequence::new(&world, 1024);
+//!     let mut s = livi::event::LV2AtomSequence::new(&features, 1024);
 //!     let play_note_data = [0x90, 0x40, 0x7f];
-//!     s.push_midi_event::<3>(1, world.midi_urid(), &play_note_data)
+//!     s.push_midi_event::<3>(1, features.midi_urid(), &play_note_data)
 //!         .unwrap();
 //!     s
 //! };
-//! let params: Vec<f32> = plugin
-//!     .ports_with_type(livi::PortType::ControlInput)
-//!     .map(|p| p.default_value)
-//!     .collect();
-//! let mut outputs = [vec![0.0; MAX_BLOCK_SIZE], vec![0.0; MAX_BLOCK_SIZE]];
-//! let ports = livi::EmptyPortConnections::new(MAX_BLOCK_SIZE)
+//!
+//! // This is where the audio data will be stored.
+//! let mut outputs = [
+//!     vec![0.0; features.max_block_length()], // For mda EPiano, this is the left channel.
+//!     vec![0.0; features.max_block_length()], // For mda EPiano, this is the right channel.
+//! ];
+//!
+//! // Set up the port configuration and run the plugin!
+//! // The results will be stored in `outputs`.
+//! let ports = livi::EmptyPortConnections::new()
 //!     .with_atom_sequence_inputs(std::iter::once(&input))
-//!     .with_audio_outputs(outputs.iter_mut().map(|output| output.as_mut_slice()))
-//!     .with_control_inputs(params.iter());
-//! unsafe { instance.run(ports).unwrap() };
+//!     .with_audio_outputs(outputs.iter_mut().map(|output| output.as_mut_slice()));
+//! unsafe { instance.run(features.max_block_length(), ports).unwrap() };
+//!
+//! // Plugins may push asynchronous works to the worker. When operating in
+//! // Realtime, `run_workers` should be run in a separate thread.
+//! std::thread::spawn(move || loop {
+//!     worker_manager.run_workers();
+//!     // Add some sleep to avoid busy looping.
+//!     // Busy looping may lead to increased CPU usage.
+//!     std::thread::sleep(std::time::Duration::from_millis(100));
+//! });
 //! ```
-use crate::error::InitializeBlockLengthError;
-use crate::features::Features;
 use log::{debug, error, info, warn};
-use std::sync::{Arc, Mutex};
+use std::sync::Arc;
 
+pub use features::worker::{Worker, WorkerManager};
+pub use features::{Features, FeaturesBuilder};
 pub use plugin::{Instance, Plugin};
 pub use port::{EmptyPortConnections, Port, PortConnections, PortCounts, PortIndex, PortType};
 
@@ -53,7 +64,6 @@ mod port;
 
 /// Contains all plugins.
 pub struct World {
-    resources: Arc<Resources>,
     livi_plugins: Vec<Plugin>,
 }
 
@@ -61,14 +71,31 @@ impl World {
     /// Create a new world that includes all plugins that are found and are
     /// supported.  Plugins that are not supported will be listed with a `warn!`
     /// message.
-    ///
-    /// # Panics
-    /// Panics if the world resources mutex could not be locked.
     #[must_use]
     pub fn new() -> World {
         World::with_plugin_predicate(|_| true)
     }
 
+    /// Creates a new world that includes only a single
+    /// plugin specified by bundle_uri.
+    /// bundle_uri must be a fully qualified URI to the bundle directory,
+    /// with the trailing slash, eg file:///usr/lib/lv2/foo.lv2/.
+    pub fn with_load_bundle(bundle_uri: &str) -> World {
+        let world = lilv::World::new();
+        let uri = world.new_uri(bundle_uri);
+        world.load_bundle(&uri);
+        let common_uris = Arc::new(CommonUris::new(&world));
+        let plugins: Vec<Plugin> = world
+            .plugins()
+            .into_iter()
+            .map(|p| Plugin::from_raw(p, common_uris.clone()))
+            .collect();
+
+        World {
+            livi_plugins: plugins,
+        }
+    }
+
     /// Creates a new world that includes all plugins that are found and return
     /// `true` for `predicate.
     #[must_use]
@@ -77,8 +104,8 @@ impl World {
         P: Fn(&Plugin) -> bool,
     {
         let world = lilv::World::with_load_all();
-        let resources = Arc::new(Resources::new(&world));
-        let supported_features = resources.features.lock().unwrap().supported_features();
+        let common_uris = Arc::new(CommonUris::new(&world));
+        let supported_features = crate::Features::supported_features();
         info!(
             "Creating World with supported features {:?}",
             supported_features
@@ -114,19 +141,19 @@ impl World {
             .filter(|p| {
                 for port in p.iter_ports() {
                     for class in port.classes() {
-                        if class != resources.input_port_uri
-                            && class != resources.output_port_uri
-                            && class != resources.audio_port_uri
-                            && class != resources.control_port_uri
-                            && class != resources.atom_port_uri
-                            && class != resources.cv_port_uri
+                        if class != common_uris.input_port_uri
+                            && class != common_uris.output_port_uri
+                            && class != common_uris.audio_port_uri
+                            && class != common_uris.control_port_uri
+                            && class != common_uris.atom_port_uri
+                            && class != common_uris.cv_port_uri
                         {
                             error!("Port class {:?} is not supported.", class);
                             return false;
                         }
                     }
-                    if !port.is_a(&resources.input_port_uri)
-                        && !port.is_a(&resources.output_port_uri)
+                    if !port.is_a(&common_uris.input_port_uri)
+                        && !port.is_a(&common_uris.output_port_uri)
                     {
                         error!(
                             "Port {:?} for plugin {} is neither an input or output.",
@@ -135,7 +162,7 @@ impl World {
                         );
                         return false;
                     }
-                    if !port.is_a(&resources.audio_port_uri) && !port.is_a(&resources.control_port_uri) && !port.is_a(&resources.atom_port_uri) && !port.is_a(&resources.cv_port_uri) {
+                    if !port.is_a(&common_uris.audio_port_uri) && !port.is_a(&common_uris.control_port_uri) && !port.is_a(&common_uris.atom_port_uri) && !port.is_a(&common_uris.cv_port_uri) {
                         error!(
                             "Port {:?}for plugin {} not a recognized data type. Supported types are Audio and Control", port, p.uri().as_str().unwrap_or("BAD_URI")
                         );
@@ -144,7 +171,7 @@ impl World {
                 }
                 true
             })
-            .map(|p| Plugin::from_raw(p, resources.clone()))
+            .map(|p| Plugin::from_raw(p, common_uris.clone()))
             .filter(|p| {
                 let keep = predicate(p);
                 if !keep {
@@ -155,35 +182,10 @@ impl World {
             .inspect(|p| info!("Found plugin {}: {}", p.name(), p.uri()))
             .collect();
         World {
-            resources,
             livi_plugins: plugins,
         }
     }
 
-    /// Get the URID of a URI. This value is only guaranteed to be valid for
-    /// instances spawned from this world. It is not guaranteed to be stable
-    /// across different runs.
-    ///
-    /// # Panics
-    /// Panics if the world resource mutex could not be locked.
-    #[must_use]
-    pub fn urid(&self, uri: &std::ffi::CStr) -> lv2_raw::LV2Urid {
-        self.resources.features.lock().unwrap().urid_map.map(uri)
-    }
-
-    /// The URID for midi events.
-    ///
-    /// # Panics
-    /// Panics if a `CStr` could not be built for the Midi URI. This behavior is
-    /// well tested and of negligible risk.
-    #[must_use]
-    pub fn midi_urid(&self) -> lv2_raw::LV2Urid {
-        self.urid(
-            std::ffi::CStr::from_bytes_with_nul(b"http://lv2plug.in/ns/ext/midi#MidiEvent\0")
-                .unwrap(),
-        )
-    }
-
     /// Iterate through all plugins.
     pub fn iter_plugins(&self) -> impl '_ + ExactSizeIterator + Iterator<Item = Plugin> {
         self.livi_plugins.iter().cloned()
@@ -201,19 +203,8 @@ impl World {
     ///
     /// # Errors
     /// Returns an error if the block lengths are invalid.
-    ///
-    /// # Panics
-    /// Panics if the world resource mutex could not be locked.
-    pub fn initialize_block_length(
-        &mut self,
-        min_block_length: usize,
-        max_block_length: usize,
-    ) -> Result<(), InitializeBlockLengthError> {
-        self.resources
-            .features
-            .lock()
-            .unwrap()
-            .initialize_block_length(min_block_length, max_block_length)
+    pub fn build_features(&self, builder: crate::features::FeaturesBuilder) -> Arc<Features> {
+        builder.build(self)
     }
 }
 
@@ -223,26 +214,26 @@ impl Default for World {
     }
 }
 
-struct Resources {
+struct CommonUris {
     input_port_uri: lilv::node::Node,
     output_port_uri: lilv::node::Node,
     control_port_uri: lilv::node::Node,
     audio_port_uri: lilv::node::Node,
     atom_port_uri: lilv::node::Node,
     cv_port_uri: lilv::node::Node,
-    features: Mutex<Features>,
+    worker_schedule_feature_uri: lilv::node::Node,
 }
 
-impl Resources {
-    fn new(world: &lilv::World) -> Resources {
-        Resources {
+impl CommonUris {
+    fn new(world: &lilv::World) -> CommonUris {
+        CommonUris {
             input_port_uri: world.new_uri("http://lv2plug.in/ns/lv2core#InputPort"),
             output_port_uri: world.new_uri("http://lv2plug.in/ns/lv2core#OutputPort"),
             control_port_uri: world.new_uri("http://lv2plug.in/ns/lv2core#ControlPort"),
             audio_port_uri: world.new_uri("http://lv2plug.in/ns/lv2core#AudioPort"),
             atom_port_uri: world.new_uri("http://lv2plug.in/ns/ext/atom#AtomPort"),
             cv_port_uri: world.new_uri("http://lv2plug.in/ns/lv2core#CVPort"),
-            features: Mutex::new(Features::new()),
+            worker_schedule_feature_uri: world.new_uri("http://lv2plug.in/ns/ext/worker#schedule"),
         }
     }
 }
@@ -262,54 +253,57 @@ mod tests {
     #[test]
     fn test_midi_urid_ok() {
         let world = World::new();
-        assert!(world.midi_urid() > 0, "midi urid is not valid");
+        let features = world.build_features(crate::features::FeaturesBuilder {
+            min_block_length: MIN_BLOCK_SIZE,
+            max_block_length: MAX_BLOCK_SIZE,
+            worker_manager: Default::default(),
+        });
+        assert!(features.midi_urid() > 0, "midi urid is not valid");
     }
 
     #[test]
     fn test_all() {
-        let mut world = World::new();
+        let world = World::new();
         let block_size = 64;
-        world
-            .initialize_block_length(block_size, block_size)
-            .unwrap();
+        let features = world.build_features(crate::features::FeaturesBuilder {
+            min_block_length: block_size,
+            max_block_length: block_size,
+            worker_manager: Default::default(),
+        });
         for plugin in world.iter_plugins() {
+            println!("Running plugin: {}", plugin.uri());
             let port_counts = *plugin.port_counts();
-            let in_params: Vec<f32> = plugin
-                .ports_with_type(PortType::ControlInput)
-                .map(|p| p.default_value)
-                .collect();
-            let mut out_params: Vec<f32> = plugin
-                .ports_with_type(PortType::ControlOutput)
-                .map(|p| p.default_value)
-                .collect();
             let audio_in = vec![0.0; port_counts.audio_inputs * block_size];
             let mut audio_out = vec![0.0; port_counts.audio_outputs * block_size];
             let cv_in = vec![0.0; port_counts.cv_inputs * block_size];
             let mut cv_out = vec![0.0; port_counts.cv_outputs * block_size];
             let play_note_data = [0x90, 0x40, 0x7f];
             let release_note_data = [0x80, 0x40, 0x00];
+            let input_events_features = features.clone();
             let input_events = (0..port_counts.atom_sequence_inputs)
                 .map(|_| {
-                    let mut seq = LV2AtomSequence::new(&world, 1024);
-                    seq.push_midi_event::<3>(4, world.midi_urid(), &play_note_data)
-                        .unwrap();
-                    seq.push_midi_event::<3>(60, world.midi_urid(), &release_note_data)
+                    let mut seq = LV2AtomSequence::new(&input_events_features, 1024);
+                    seq.push_midi_event::<3>(4, input_events_features.midi_urid(), &play_note_data)
                         .unwrap();
+                    seq.push_midi_event::<3>(
+                        60,
+                        input_events_features.midi_urid(),
+                        &release_note_data,
+                    )
+                    .unwrap();
                     seq
                 })
                 .collect::<Vec<_>>();
+            let output_events_features = features.clone();
             let mut output_events = (0..port_counts.atom_sequence_outputs)
-                .map(|_| LV2AtomSequence::new(&world, 1024))
+                .map(|_| LV2AtomSequence::new(&output_events_features, 1024))
                 .collect::<Vec<_>>();
             let mut instance = unsafe {
                 plugin
-                    .instantiate(SAMPLE_RATE)
+                    .instantiate(features.clone(), SAMPLE_RATE)
                     .expect("Could not instantiate plugin.")
             };
             let ports = PortConnections {
-                sample_count: block_size,
-                control_inputs: in_params.iter(),
-                control_outputs: out_params.iter_mut(),
                 audio_inputs: audio_in
                     .chunks_exact(block_size)
                     .take(port_counts.audio_inputs),
@@ -325,7 +319,7 @@ mod tests {
             };
             unsafe {
                 assert_eq!(
-                    instance.run(ports),
+                    instance.run(block_size, ports),
                     Ok(()),
                     "Failed on run {} with plugin: {}",
                     block_size,
@@ -337,10 +331,7 @@ mod tests {
 
     #[test]
     fn test_mda_epiano() {
-        let mut world = World::new();
-        world
-            .initialize_block_length(MIN_BLOCK_SIZE, MAX_BLOCK_SIZE)
-            .unwrap();
+        let world = World::new();
         let plugin = world
             // Electric Piano instrument.
             .plugin_by_uri("http://drobilla.net/plugins/mda/EPiano")
@@ -358,9 +349,14 @@ mod tests {
                 cv_outputs: 0,
             }
         );
+        let features = world.build_features(FeaturesBuilder {
+            min_block_length: MIN_BLOCK_SIZE,
+            max_block_length: MAX_BLOCK_SIZE,
+            worker_manager: Default::default(),
+        });
         let mut instance = unsafe {
             plugin
-                .instantiate(SAMPLE_RATE)
+                .instantiate(features.clone(), SAMPLE_RATE)
                 .expect("Could not instantiate plugin.")
         };
         assert_eq!(
@@ -377,23 +373,18 @@ mod tests {
             }
         );
         let input = {
-            let mut s = LV2AtomSequence::new(&world, 1024);
+            let mut s = LV2AtomSequence::new(&features, 1024);
             let play_note_data = [0x90, 0x40, 0x7f];
-            s.push_midi_event::<3>(1, world.midi_urid(), &play_note_data)
+            s.push_midi_event::<3>(1, features.midi_urid(), &play_note_data)
                 .unwrap();
             s
         };
-        let params: Vec<f32> = plugin
-            .ports_with_type(PortType::ControlInput)
-            .map(|p| p.default_value)
-            .collect();
         let mut outputs = [vec![0.0; MAX_BLOCK_SIZE], vec![0.0; MAX_BLOCK_SIZE]];
         for block_size in MIN_BLOCK_SIZE..MAX_BLOCK_SIZE {
-            let ports = EmptyPortConnections::new(block_size)
+            let ports = EmptyPortConnections::new()
                 .with_atom_sequence_inputs(std::iter::once(&input))
-                .with_audio_outputs(outputs.iter_mut().map(|output| output.as_mut_slice()))
-                .with_control_inputs(params.iter());
-            unsafe { instance.run(ports).unwrap() };
+                .with_audio_outputs(outputs.iter_mut().map(|output| output.as_mut_slice()));
+            unsafe { instance.run(block_size, ports).unwrap() };
         }
         for output in outputs.iter_mut() {
             assert!(
@@ -405,15 +396,12 @@ mod tests {
 
     #[test]
     fn test_fifths() {
-        let mut world = World::new();
+        let world = World::new();
         let block_size = 128;
-        world
-            .initialize_block_length(block_size, block_size)
-            .unwrap();
         let plugin = world
             // Takes a midi and adds the fifth of every note.
             .plugin_by_uri("http://lv2plug.in/plugins/eg-fifths")
-            .expect("Plugin not found.");
+            .expect("Plugin http://lv2plug.in/plugins/eg-fifths not found.");
         assert_eq!(
             *plugin.port_counts(),
             PortCounts {
@@ -427,9 +415,14 @@ mod tests {
                 cv_outputs: 0,
             }
         );
+        let features = world.build_features(FeaturesBuilder {
+            min_block_length: block_size,
+            max_block_length: block_size,
+            worker_manager: Default::default(),
+        });
         let mut instance = unsafe {
             plugin
-                .instantiate(SAMPLE_RATE)
+                .instantiate(features.clone(), SAMPLE_RATE)
                 .expect("Could not instantiate plugin.")
         };
 
@@ -439,25 +432,25 @@ mod tests {
         let release_c4 = [0x80, 0x3C, 0x00];
         let release_g4 = [0x80, 0x43, 0x00];
 
-        let mut input = LV2AtomSequence::new(&world, 1024);
+        let mut input = LV2AtomSequence::new(&features, 1024);
         input
-            .push_midi_event::<3>(1, world.midi_urid(), &play_c4)
+            .push_midi_event::<3>(1, features.midi_urid(), &play_c4)
             .unwrap();
         input
-            .push_midi_event::<3>(10, world.midi_urid(), &release_c4)
+            .push_midi_event::<3>(10, features.midi_urid(), &release_c4)
             .unwrap();
 
-        let mut output = LV2AtomSequence::new(&world, 1024);
+        let mut output = LV2AtomSequence::new(&features, 1024);
         // This note should be cleared from the output by the LV2 plugin.
         output
-            .push_midi_event::<3>(1, world.midi_urid(), &play_c3)
+            .push_midi_event::<3>(1, features.midi_urid(), &play_c3)
             .unwrap();
 
         for _ in 0..10 {
-            let ports = EmptyPortConnections::new(block_size)
+            let ports = EmptyPortConnections::new()
                 .with_atom_sequence_inputs(std::iter::once(&input))
                 .with_atom_sequence_outputs(std::iter::once(&mut output));
-            unsafe { instance.run(ports).unwrap() };
+            unsafe { instance.run(block_size, ports).unwrap() };
         }
 
         let got = output
@@ -498,23 +491,17 @@ mod tests {
 
     #[test]
     fn test_supported_features() {
-        let want: HashSet<String> = [
+        let supported_features = Features::supported_features();
+
+        assert!(supported_features.contains("http://lv2plug.in/ns/ext/urid#map"));
+
+        let want = HashSet::from([
             "http://lv2plug.in/ns/ext/urid#map",
             "http://lv2plug.in/ns/ext/urid#unmap",
             "http://lv2plug.in/ns/ext/options#options",
             "http://lv2plug.in/ns/ext/buf-size#boundedBlockLength",
-        ]
-        .iter()
-        .map(|s| s.to_string())
-        .collect();
-        assert_eq!(
-            want,
-            World::new()
-                .resources
-                .features
-                .lock()
-                .unwrap()
-                .supported_features()
-        );
+            "http://lv2plug.in/ns/ext/worker#schedule",
+        ]);
+        assert_eq!(want, supported_features);
     }
 }
